/*
 * lxdash-d.c.c
 * Copyright (C) 2008-2015, 2017, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stddef.h>

#include <copy.h>
#include <d_fset.h>
#include <e4.h>
#include <fd.h>
#include <lxcast.h>
#include <lxdash-close.h>
#include <lxdash-inter.h>
#include <lxdash-types.h>
#include <nt.h>

#define I_LINE(t)			(*((X1f4_E4_C_USER *) (t)))
#define I_MODE(t)			(*((X1f4_E4_C_MODE *) (t)))
#define I_USER(t)			(*((X1f4_E4_C_USER *) (t)))

#define l_USER(e, output) \
    {									      \
	X1f4_E4_C_USER *l;						      \
									      \
	l = (void *) (output);						      \
	*l = (e);							      \
    }

int
_libx1f4i0_lxdash_c_derived(void *context, void *output, void **input)
{
    int status, type;
    struct dxdash_type *dxdash_data;
    struct lxdash_type *lxdash_data;
    struct lxtext_type *lxtext_data, *lxtext_link;
    struct x1f4_dxcast_type *dxcast_data;
    unsigned miss;

    dxdash_data = I_LINE(input[0]);

    lxdash_data = context;

    dxcast_data = I_USER(input[2]);

    type = dxcast_data->type;

    lxtext_link = lxdash_data->link_f.text;

    lxtext_data = lxtext_link;

    close_frame(miss, lxtext_data);

    do {
	int (*copy)(void *, void *, void **, const void *,
		    const struct x1f4_nodelink_type *),
	    (*node)(void *, void **, void *,
		    const struct x1f4_nodelink_type *);
	struct effect_type effect, *effect_data;
	unsigned link;
	void *data, *side;

	if (miss) {
	    if (lxtext_data->lxtype.flags & X1f4_LX_LINK_ACCESS) {
		node = lxtext_data->lxtype.node;
		copy = NULL;
	    } else {
		node = NULL;
		copy = lxtext_data->lxtype.copy;
		if (copy) {
		    status = copy
			(lxtext_data->lxtype.context, NULL, &side,
			 I_USER(&dxcast_data->data), lock_excase(dxdash_data));
		    if (status) {
			break;
		    }
		}
	    }
	} else {
	    copy = NULL;
	    node = NULL;
	}

	effect.mode = I_MODE(input[1]);
	status = x1f4_post_lxfset(dxdash_data->dash, &effect, &data);
	if (status) {
	    if (status ^ X1f4_BQFSET_EVER_MATCH) {
		status = _libx1f4i0_lxdash_stat_line(lxdash_data, 0, status);

		if (copy) {
		    lxtext_data->lxtype.free
			(lxtext_data->lxtype.context, side,
			 NULL, lock_excase(dxdash_data));
		}

		break;
	    } else {
		unsigned link;

		effect_data = data;

		link = effect_data->link;

		if (~link) {
		    int (*free)(void *, void *, struct excase_type **,
				const struct excase_type *);
		    struct lxtext_type *lxtext_slip;

		    lxtext_slip = lxtext_link + link;

		    free = lxtext_slip->lxtype.free;
		    if (free) {
			status = free
			    (lxtext_slip->lxtype.context,
			     I_USER(&effect_data->fxdata), NULL,
			     lock_excase(dxdash_data));
			if (status) {
			    if (copy) {
				lxtext_data->lxtype.free
				    (lxtext_data->lxtype.context, side,
				     NULL, lock_excase(dxdash_data));
			    }

			    break;
			}
		    } else {
			status = 0;
		    }
		} else {
		    status = 0;
		}
	    }
	} else {
	    dxdash_data->size++;

	    effect_data = data;

	    effect_data->mode = effect.mode;
	}

	effect_data->type = type;

	link = miss ? lxtext_data - lxtext_link : ~0;

	if (node) {
	    status = node
		(lxtext_data->lxtype.context, (void *) &effect_data->fxdata,
		 I_USER(&dxcast_data->data), lock_excase(dxdash_data));
	    if (status) {
		link = ~0;
	    }
	} else {
	    if (copy) {
		l_USER(side, &effect_data->fxdata);
	    } else {
		copy_miss(&effect_data->fxdata, type, &dxcast_data->data);
	    }
	}

	effect_data->link = link;

	if (status) {
	} else {
	    if (copy || node) {
		status = _libx1f4i0_lxrail_post_data
		    (context, output, link, type, &effect_data->fxdata);
	    } else {
		l_USER(dxcast_data, output);
	    }
	}
    } while (0);

    return status;
}
