/*
 * lxdecq-types.h
 * Copyright (C) 2008-2017, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __LIBx1f4_lxdecq_types_H__
#define __LIBx1f4_lxdecq_types_H__

#include <config.h>

#include <lxdecq-config.h>

#include <dt.h>
#include <e4.h>
#include <fd.h>
#include <frame.h>
#include <integral_q.h>
#include <ln.h>
#include <lxrail.h>
#include <nl.h>
#include <track.h>
#include <trans.h>

#define excase_type			x1f4_nodelink_type
#define exlist_type			x1f4_nodeline_type

#define cxlist_type			x1f4_leadnode_type

#define here_NULL			((void *) 0)

#define here_offsetof(type, ceil) \
    ((integral_q) (&((type *) here_NULL)->ceil))

#define lock_excase(dxlist_text) \
    ((struct excase_type *) &lock_cxlist((dxlist_text))->link)

#define lock_exlist(dxlist_text) \
    ((struct exlist_type *) &lock_cxlist((dxlist_text))->line)

#define lock_cxlist(dxlist_text) \
    ((struct cxlist_type *)						      \
     ((char *) (dxlist_text) - sizeof(struct cxlist_type)))

#define near_dxlist(excase_text) \
    ((struct dxlist_type *)						      \
     ((char *) near_cxlist((excase_text)) + sizeof(struct cxlist_type)))

#define near_exlist(excase_text) \
    ((struct exlist_type *) &near_cxlist((excase_text))->line)

#define near_cxlist(excase_text) \
    ((struct cxlist_type *)						      \
     ((char *) (excase_text) - here_offsetof(struct cxlist_type, link)))

#define pick_dxlist(fxlist_text) \
    ((struct dxlist_type *)						      \
     ((char *) (fxlist_text) + sizeof(struct cxlist_type)))

#define post_dxlist(dxlist_text) \
    ((struct dxlist_type *) (dxlist_text) + 1)

#define __struct_dxlist_type_alloc_size__ \
    sizeof(struct cxlist_type) + sizeof(struct dxlist_type)

typedef struct cxbend_type {
    int bend, (*pick)(void *, void **), (*push)(void *, void **),
        (*roll)(void *), (*slip)(void *);
    void *list;
} cxbend_type;

typedef struct dxlist_type {
    unsigned size;
    void *list;
} dxlist_type;

typedef struct lxdecq_type {
    struct lxrail_type ever_r;
    struct {
        struct cxbend_type cxbend[2];
    } rule_f;
} lxdecq_type;

typedef struct lxmiss_type {
    union {
	unsigned miss;
	void *data;
    } data;
} lxmiss_type;

typedef struct screen_type {
    struct x1f4_track_type lock_e;
    struct {
	const struct x1f4_eelookup_type *eelookup_data;
    } lock_g;
    struct {
	int type;
    } link_l;
    struct x1f4_trans_type lock_m;
    struct {
	int (*call)(void *, struct x1f4_nodelink_type *),
	    (*fine)(void *, struct x1f4_nodelink_type **,
		    struct x1f4_nodelink_type *),
	    (*miss)(void *, struct x1f4_nodelink_type **,
		    struct x1f4_nodelink_type *);
	void *text;
    } lock_r;
    struct {
	unsigned miss;
	const void *data;
    } lock_t;
    struct x1f4_frame_type lock_v;
    struct x1f4_trans_type lock_w;
} screen_type;

#endif				/* __LIBx1f4_lxdecq_types_H__ */
