/*
 * lxline-m.o.c
 * Copyright (C) 2008-2017, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <lxline-config.h>

#include <e4.h>
#include <l_note.h>
#include <lxline-drive.h>
#include <lxline-inter.h>
#include <lxline-types.h>

#define I_LINE(t)			(*((X1f4_E4_C_USER *) (t)))
#define I_MODE(t)			(*((X1f4_E4_C_MODE *) (t)))
#define I_TEXT(t)			(*((X1f4_E4_C_TEXT *) (t)))

#define l_MODE(e, output) \
    {									      \
	X1f4_E4_C_MODE *l;						      \
									      \
	l = (void *) (output);						      \
	*l = (e);							      \
    }

#if __HAVE_SHUT_SFNOTE__
static int shut_note(void *, void *);
#endif				/* __HAVE_SHUT_SFNOTE__ */

#if __HAVE_SHUT_SFNOTE__
static int
shut_note(void *text, void *data)
{
    int status = 0;
    unsigned type;

    type = effect_type(data);
    if (type != X1f4_E4_MODE) {
	status = 1;
    } else {
	l_MODE(I_MODE(byte(data) + 8), *(void **) text);
    }

    return status;
}
#endif				/* __HAVE_SHUT_SFNOTE__ */


int
_libx1f4i0_lxline_o_class_m(void *context, void *output, void **input)
{
    X1f4_E4_C_TEXT fail;
    int status;
    struct dxline_type *dxline_data;
    struct lxline_type *lxline_data;
    struct lxtext_type *lxtext_data;
#if __HAVE_SHUT_SFNOTE__
#else
    void *data;
#endif				/* __HAVE_SHUT_SFNOTE__ */

    lxtext_data = context;

    lxline_data = lxtext_data->text;

    dxline_data = I_LINE(input[1]);

    fail = I_TEXT(input[2]);

#if __HAVE_SHUT_SFNOTE__
    status = _libx1f4i0_shut_sfnote(dxline_data->line, fail, input, shut_note);
    if (status) {
	if (status == X1f4_SFNOTE_MISS_CLASS
	    || status == X1f4_SFNOTE_CALL_ERROR) {
	    status = 0;

	    l_MODE(0, output);
	} else {
	    status = _libx1f4i0_lxline_stat_line(lxline_data, 0, status);
	}
    } else {
	dxline_data->size--;

	if (1) {
	    l_MODE(1, output);
	}
    }
#else
    status = _libx1f4i0_find_sfnote(dxline_data->line, fail, &data);
    if (status) {
	unsigned type;

	type = effect_type(data);
	if (type != X1f4_E4_MODE) {
	    l_MODE(0, output);
	    if (1) {
		status = 0;
	    }
	} else {
	    X1f4_E4_C_MODE fast;

	    fast = I_MODE(byte(data) + 8);

	    status = _libx1f4i0_pass_sfnote(dxline_data->line, fail);
	    if (status) {
		status = _libx1f4i0_lxline_stat_line(lxline_data, 0, status);
	    } else {
		dxline_data->size--;

		l_MODE(1, output);

		if (0) {
		} else {
		    l_MODE(fast, input[0]);
		}
	    }
	}
    } else {
	l_MODE(0, output);
    }
#endif				/* __HAVE_SHUT_SFNOTE__ */

    return status;
}
