/*
 * lxlist-p.o.c
 * Copyright (C) 2011-2017, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <e4.h>
#include <lxcast.h>
#include <lxlist-inter.h>
#include <lxlist-types.h>

#define I_LIST(t)			(*((X1f4_E4_C_USER *) (t)))
#define I_MODE(t)			(*((X1f4_E4_C_MODE *) (t)))
#define I_USER(t)			(*((X1f4_E4_C_USER *) (t)))

extern const void *const x1f4_a1_walk_link;

int
_libx1f4i0_lxlist_o_pending(void *context, void *output, void **input)
{
    int status;

    do {
	X1f4_E4_C_MODE head, lock, tail;
	struct dxlist_type *dxlist_data;
	struct x1f4_dxcast_type *dxcast_data;
	const struct x1f4_function_type *function_text;
	struct x1f4_linetext_type *linetext_data;
	unsigned class, count, grade;
	void **track, **used;

	if (output == x1f4_a1_walk_link) {
	    track = input[1];
	} else {
	    track = input;
	}

	used = track[-1];

	dxcast_data = I_USER(track[1]);

	if (X1f4_E4_CALL < dxcast_data->type) {
	} else {
	    status = _libx1f4i0_lxlist_stat_roll(context, "l_ocall", 7);
	    break;
	}

	linetext_data = I_USER(&dxcast_data->data);

	if (linetext_data->function.function) {
	} else {
	    status = _libx1f4i0_lxlist_stat_zero(context, "l_ocall", 7);
	    break;
	}

	dxlist_data = I_LIST(track[0]);

	count = dxlist_data->size;

	lock = I_MODE(track[2]);

	function_text = used[X1f4_E4_RENDITION];

	class = function_text->count - 5;

	grade = class + 1;

	if (lock < -(int) grade || !(lock < (int) grade)) {
	    status = _libx1f4i0_lxlist_stat_land(context, grade, lock);

	    break;
	}

	head = I_MODE(track[3]);

	if (head < -(X1f4_E4_C_MODE) (int) count
	    || !(head < (X1f4_E4_C_MODE) (int) count)) {
	    status = _libx1f4i0_lxlist_stat_deck(context, count, head);

	    break;
	}

	tail = I_MODE(track[4]);

	if (tail < -(X1f4_E4_C_MODE) (int) count
	    || !(tail < (X1f4_E4_C_MODE) (int) count)) {
	    status = _libx1f4i0_lxlist_stat_deck(context, count, tail);

	    break;
	}

	if (lock < (X1f4_E4_C_MODE) 0) {
	    lock += grade;
	}

	if (head < (X1f4_E4_C_MODE) 0) {
	    head += count;
	}
	if (tail < (X1f4_E4_C_MODE) 0) {
	    tail += count;
	}

	if (tail < head) {
	    count = count - head + tail + 1;
	} else {
	    count = tail + 1 - head;
	}

	status = _libx1f4i0_lxlist_roll_list
	    (context, output, input, track, linetext_data, count,
	     function_text, 5, lock, dxlist_data, head, tail);
    } while (0);

    return status;
}
