/*
 * lxlock-a.e.c
 * Copyright (C) 2008-2010, 2013-2017, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#if defined HAVE_LIBx1f4i0
# include <libx1f4i0.h>
#endif				/* HAVE_LIBx1f4i0 */

#include <er.h>
#include <exerrors.h>
#include <lxcall.h>
#include <lxlock.h>

#define lxlock(lxlock) \
    ((struct lxlock_type *) (lxlock))

int
_libx1f4i0_lxlock_stat_free(void *lxlock, int shadow)
{
    LINE_EEER(lxlock(lxlock)->diff_e);

    PUSH_DATA("cannot free memory", 18);

    POST_EEER();

    return _libx1f4i0_lxcall_land_slip(X1f4_EX_CRITICAL, shadow);
}


int
_libx1f4i0_lxlock_stat_land(void *lxlock, unsigned size, long mode)
{
    LINE_EEER(lxlock(lxlock)->diff_e);

    PUSH_DATA("out of valid range (-", 21);
    PUSH_BILL(size);
    PUSH_DATA(" .. ", 4);
    PUSH_BILL(size - 1);
    PUSH_DATA(") insertion position (", 22);
    PUSH_LONG(mode);
    PUSH_DATA(")", 1);

    POST_EEER();

    return X1f4_EX_CAN_CONTINUE;
}


int
_libx1f4i0_lxlock_stat_link(void *lxlock)
{
    LINE_EEER(lxlock(lxlock)->diff_e);

    PUSH_DATA("cannot allocate memory", 22);

    POST_EEER();

    return X1f4_EX_CRITICAL;
}


int
_libx1f4i0_lxlock_stat_mode(void *lxlock)
{
    LINE_EEER(lxlock(lxlock)->diff_e);

    PUSH_DATA("cannot reallocate memory", 24);

    POST_EEER();

    return X1f4_EX_CRITICAL;
}


int
_libx1f4i0_lxlock_stat_roll(void *lxlock, const char *call, unsigned size)
{
    LINE_EEER(lxlock(lxlock)->diff_e);

    PUSH_DATA("non function pointer first argument for function `", 50);
    PUSH_DATA(call, size);
    PUSH_DATA("'", 1);

    POST_EEER();

    return X1f4_EX_CAN_CONTINUE;
}


int
_libx1f4i0_lxlock_stat_zero(void *lxlock, const char *call, unsigned size)
{
    LINE_EEER(lxlock(lxlock)->diff_e);

    PUSH_DATA(call, size);
    PUSH_DATA(": null function pointer", 23);

    POST_EEER();

    return X1f4_EX_CAN_CONTINUE;
}
