/*
 * lxpoll-i.s.c
 * Copyright (C) 2016, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <d_e101.h>
#include <e4.h>
#include <lxpoll-defs.h>
#include <lxpoll-types.h>

/*
 * returning control to application for step by step execution is possible,
 * though the BACK LEAD method is somewhat inconvenient as it makes use of a
 * _LEAD_CALL function and thus it requires the application to recognize
 * (artificial) _LEAD_CALL control returns.  and it doesn't work for
 * _slip_program, as the function is not equipped for _LEAD_CALL processing.
 */

#define __BACK_LEAD__			0

#define dxpoll(dxpoll) \
    ((struct dxpoll_type *) (dxpoll))
#define lxpoll(lxpoll) \
    ((struct lxpoll_type *) (lxpoll))

int
_libx1f4i0_lxpoll_bail_poll(void *dxpoll, void *lxpoll)
{
    int (*call)(void *), status = 0;

    call = lxpoll(lxpoll)->link_u.call;
    if (call) {
	if (*lxpoll(lxpoll)->link_u.hold) {
#if __BACK_LEAD__
	    struct x1f4_dispatch_type *dispatch_data;

	    /*
	     * the _dxpoll_ object is to be passed through
	     * _Mylene101_line_device_.
	     */
	    dispatch_data = (void *) (dxpoll(dxpoll) + 1);
	    if (dispatch_data->back) {
		void **track;

		track = a LINE_LEAD long array of void pointers that lives for
		    long enough (could be allocated with the dxpoll object);
		track = track + LINE_LEAD;

		dxpoll(dxpoll)->bits |= POLL_FAILED;

		track[NODE_LEAD] = (void *) 1;

		dispatch_data->transfer.function_data =
		    a do nothing _LEAD_CALL function;
		dispatch_data->transfer.input = track;

		Mylene101_fail_device(dxpoll(dxpoll)->device);
	    } else {
#endif				/* __BACK_LEAD__ */
		status = call(lxpoll(lxpoll)->link_u.text);
#if __BACK_LEAD__
	    }
#endif				/* __BACK_LEAD__ */
	}
    }

    return status;
}
