/*
 * lxtrap-f.c.c
 * Copyright (C) 2011-2015, 2017, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <e4.h>
#include <lxcall.h>
#include <lxcast.h>
#include <lxtrap-inter.h>
#include <lxtrap-types.h>

#define POST_BITS			X1f4_LXCALL_POST_WAYS

#define I_MODE(t)			(*((X1f4_E4_C_MODE *) (t)))
#define I_USER(t)			(*((X1f4_E4_C_USER *) (t)))

typedef struct slip_type {
    int (*call)(void *, int, void *, struct x1f4_dispatch_type **);
    struct x1f4_dispatch_type *dispatch_data;
    unsigned loop;
    void *back, *rendition, **used;
} slip_type;

extern const void *const x1f4_a1_walk_link;

static int part_call(void *, int, void *, struct x1f4_dispatch_type **);
static int pick_call(struct x1f4_linetext_type *, unsigned, void *, void **,
		     struct lxtrap_type *);
static int slip_call(struct x1f4_linetext_type *, unsigned, void **, void *,
		     void **, struct lxtrap_type *);

static int
part_call(void *back, int status, void *output,
	  struct x1f4_dispatch_type **dispatch)
{
    do {
	unsigned slip;
	struct slip_type *call_data;

	if (status) {
	    break;
	}

	call_data = back;

	slip = call_data->loop;
	if (slip) {
	    struct x1f4_dispatch_type *dispatch_data;

	    dispatch_data = call_data->dispatch_data;

	    call_data->loop = slip - 1;

	    dispatch_data->back = call_data;
	    dispatch_data->call = part_call;

	    *dispatch = dispatch_data;

	    call_data->used[X1f4_E4_RENDITION] = call_data->rendition;

	    dispatch_data->transfer.input[-1] = call_data->used;
	} else {
	    status = call_data->call(call_data->back, 0, output, dispatch);
	}
    } while (0);

    return status;
}


/*
 * likely, used[X1f4_E4_RENDITION] is set twice for the first call and
 * uselessly for error.
 */
static int
pick_call(struct x1f4_linetext_type *linetext_data, unsigned loop,
	  void *output, void **input, struct lxtrap_type *lxtrap_data)
{
    int status;
    void *clear[3], **used;

    clear[0] = output;
    clear[1] = input;

    used = input[-1];

    status = _libx1f4i0_lxcall_post_call
	(linetext_data, used[X1f4_E4_RENDITION], (void *) x1f4_a1_walk_link, 2,
	 clear, POST_BITS, &lxtrap_data->link_v, &lxtrap_data->link_p,
	 &lxtrap_data->link_q, &lxtrap_data->link_e,
	 lxtrap_data->trap_e.eelookup_data);
    if (status) {
    } else {
	struct x1f4_dispatch_type *dispatch_data;
	struct x1f4_transfer_type *transfer_data;
	void *rendition;

	dispatch_data = clear[2];

	transfer_data = &dispatch_data->transfer;

	rendition = used[X1f4_E4_RENDITION];

	for (; loop; loop--) {
	    used[X1f4_E4_RENDITION] = rendition;

	    transfer_data->input[-1] = used;

	    status = dispatch_data->transfer.function_data->function
		(transfer_data->context, transfer_data->output,
		 transfer_data->input);
	    if (status) {
		break;
	    }
	}

	used[X1f4_E4_RENDITION] = rendition;

	if (status) {
	} else {
	    status = dispatch_data->call
		(dispatch_data->back, 0, output, &dispatch_data);
	}
    }

    return status;
}


static int
slip_call(struct x1f4_linetext_type *linetext_data, unsigned loop,
	  void **track, void *output, void **input,
	  struct lxtrap_type *lxtrap_data)
{
    int status;
    void **used;

    used = track[-1];

    status = _libx1f4i0_lxcall_post_call
	(linetext_data, used[X1f4_E4_RENDITION], output, 2, input, POST_BITS,
	 &lxtrap_data->link_v, &lxtrap_data->link_p, &lxtrap_data->link_q,
	 &lxtrap_data->link_e, lxtrap_data->trap_e.eelookup_data);
    if (status) {
    } else {
	loop--;
	if (loop) {
	    void *call;

	    status = lxtrap_data->link_v.link
		(lxtrap_data->link_v.data, &call, sizeof(struct slip_type));
	    if (status) {
		status = _libx1f4i0_lxtrap_stat_link(lxtrap_data);
	    } else {
		struct slip_type *call_data;
		struct x1f4_dispatch_type *dispatch_data;

		dispatch_data = input[2];

		call_data = call;

		call_data->back = dispatch_data->back;
		call_data->call = dispatch_data->call;
		call_data->loop = loop;

		call_data->used = used;
		call_data->rendition = used[X1f4_E4_RENDITION];

		dispatch_data->back = call_data;
		dispatch_data->call = part_call;

		call_data->dispatch_data = dispatch_data;
	    }
	}
    }

    return status;
}


int
_libx1f4i0_lxtrap_c_forward(void *context, void *output, void **input)
{
    int status;
    struct lxtrap_type *lxtrap_data;
    struct x1f4_dxcast_type *dxcast_data;
    void **track;

    lxtrap_data = context;

    if (output == x1f4_a1_walk_link) {
	track = input[1];
    } else {
	track = input;
    }

    dxcast_data = I_USER(track[1]);

    if (X1f4_E4_CALL < dxcast_data->type) {
	struct x1f4_linetext_type *linetext_data;

	linetext_data = I_USER(&dxcast_data->data);
	if (linetext_data->function.function) {
	    X1f4_E4_C_MODE mode;

	    mode = I_MODE(track[0]);

	    if (0 < mode) {
		if (output == x1f4_a1_walk_link) {
		    status = slip_call
			(linetext_data, mode, track, output, input,
			 lxtrap_data);
		} else {
		    status = pick_call
			(linetext_data, mode, output, input, lxtrap_data);
		}
	    } else {
		status = 0;

		if (output == x1f4_a1_walk_link) {
		    input[2] = NULL;
		}
	    }
	} else {
	    status = _libx1f4i0_lxtrap_stat_zero(lxtrap_data, "call_n", 6);
	}
    } else {
	status = _libx1f4i0_lxtrap_stat_call(lxtrap_data, 1, "call_n", 6);
    }

    return status;
}
