/*
 * lxtrap-i.y.c
 * Copyright (C) 2011-2015, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stddef.h>
#include <string.h>

#include <exerrors.h>
#include <lxtrap-types.h>
#include <track.h>

extern const char *const x1f4_c1_empty_string;

static int line_flat(void *, int);
static int post_flat(void *);
static int push_flat(void *, const char *, unsigned);

const struct x1f4_track_type _libx1f4i0_stored_track = {
/* *INDENT-OFF* */
    line_flat,
    post_flat,
    push_flat,
    NULL
/* *INDENT-ON* */
};

static int
line_flat(void *fxcase, int direct)
{
    ((struct fxcase_type *) fxcase)->missed = 0;

    return 0;
}


static int
post_flat(void *fxcase)
{
    return 0;
}


static int
push_flat(void *fxcase, const char *data, unsigned size)
{
    int status;

    if (size) {
	struct fxcase_type *fxcase_data;

	fxcase_data = fxcase;

	if (fxcase_data->text == x1f4_c1_empty_string) {
	    unsigned trip;
	    void *miss;

	    trip = size;
	    if (trip < 96) {
		trip = 96;
	    }

	    status = fxcase_data->trans.link
		(fxcase_data->trans.data, &miss, trip + 1);
	    if (status) {
		status = X1f4_EX_CRITICAL;
	    } else {
		X1f4_E4_C_TEXT text;

		text = miss;
		memcpy(text, data, size);
		text[size] = 0;

		fxcase_data->size = size;
		fxcase_data->text = text;
		fxcase_data->trip = trip;
	    }
	} else {
	    X1f4_E4_C_TEXT text;
	    unsigned call, post;

	    text = fxcase_data->text;

	    post = fxcase_data->size;

	    call = post + size;
	    if (fxcase_data->trip < call) {
		unsigned trip;
		void *miss;

		trip = call;

		trip |= trip >> 001;
		trip |= trip >> 002;
		trip |= trip >> 004;
		trip |= trip >> 010;
		trip |= trip >> 020;

		miss = text;
		status = fxcase_data->trans.mode
		    (fxcase_data->trans.data, &miss, trip + 1);
		if (status) {
		    status = X1f4_EX_CRITICAL;
		} else {
		    text = miss;
		    text += post;
		    memcpy(text, data, size);
		    text[size] = 0;

		    fxcase_data->size = call;
		    fxcase_data->text = miss;
		    fxcase_data->trip = trip;
		}
	    } else {
		status = 0;

		text += post;
		memcpy(text, data, size);
		text[size] = 0;

		fxcase_data->size = call;
	    }
	}
    } else {
	status = 0;
    }

    return status;
}
